/* eslint-disable no-constant-condition */
import { promises as fs, constants } from 'fs';
import { Client } from '.';

type JSONData = [{key: string, value: any}?];


/**
 * Persistant local JSON-based storage.
 * Auto-locking system to prevent corrupted data.
 * @author Matthew <matthew@staff.libraryofcode.org>
 */
export default class LocalStorage {
  private client: Client;

  protected storagePath: string;

  private locked: boolean = false;

  constructor(client: Client, storagePath = `${__dirname}/../../localstorage`) {
    this.client = client;
    this.storagePath = storagePath;
    this.init();
  }

  private async init() {
    try {
      await fs.access(`${this.storagePath}/1.json`, constants.F_OK);
    } catch {
      const setup = [];
      await fs.writeFile(`${this.storagePath}/1.json`, JSON.stringify(setup), { encoding: 'utf8' });
    }
  }

  /**
   * Retrieves one data from the store.
   * If the store has multiple entries for the same key, this function will only return the first entry.
   * ```ts
   * await LocalStorage.get<type>('data-key');
   * ```
   * @param key The key for the data entry.
   */
  public async get<T>(key: string): Promise<T> {
    while (true) {
      if (!this.locked) break;
    }
    this.locked = true;

    const file = await fs.readFile(`${this.storagePath}/1.json`, { encoding: 'utf8' });
    this.locked = false;
    const json: JSONData = JSON.parse(file);
    const result = json.filter((data) => data.key === key);
    if (!result[0]) return null;
    return result[0].value;
  }

  /**
   * Retrieves multiple data keys/values from the store.
   * This function will return all of the values matching the key you provided exactly. Use `LocalStorage.get();` if possible.
   * ```ts
   * await LocalStorage.get<type>('data-key');
   * @param key The key for the data entry.
   */
  public async getMany<T>(key: string): Promise<{key: string, value: T}[]> {
    while (true) {
      if (!this.locked) break;
    }
    this.locked = true;

    const file = await fs.readFile(`${this.storagePath}/1.json`, { encoding: 'utf8' });
    this.locked = false;
    const json: JSONData = JSON.parse(file);
    const result = json.filter((data) => data.key === key);
    if (result.length < 1) return null;
    return result;
  }

  /**
   * Sets a key/value pair and creates a new data entry.
   * @param key The key for the data entry.
   * @param value The value for the data entry, can be anything that is valid JSON.
   * @param options.override [DEPRECATED] By default, this function will error if the key you're trying to set already exists. Set this option to true to override that setting.
   * ```ts
   * await LocalStorage.set('data-key', 'test');
   * ```
   */
  public async set(key: string, value: any): Promise<void> {
    while (true) {
      if (!this.locked) break;
    }
    this.locked = true;

    const file = await fs.readFile(`${this.storagePath}/1.json`, { encoding: 'utf8' });
    const json: JSONData = JSON.parse(file);
    json.push({ key, value });
    await fs.writeFile(`${this.storagePath}/1.json`, JSON.stringify(json), { encoding: 'utf8' });
    this.locked = false;
  }

  /**
   * Deletes the data for the specified key.
   * **Warning:** This function will delete ALL matching entries.
   * ```ts
   * await LocalStorage.del('data-key');
   * ```
   * @param key The key for the data entry.
   */
  public async del(key: string): Promise<void> {
    while (true) {
      if (!this.locked) break;
    }
    this.locked = true;

    const file = await fs.readFile(`${this.storagePath}/1.json`, { encoding: 'utf8' });
    const json: JSONData = JSON.parse(file);
    const filtered = json.filter((data) => data.key !== key);
    await fs.writeFile(`${this.storagePath}/1.json`, JSON.stringify(filtered), { encoding: 'utf8' });
    this.locked = false;
  }
}
